@@ -14,11 +14,14 @@ use serde::Deserialize;
 use smallbox::smallbox;
 use tokio::time::Instant;
 
-use elfo_utils::CachePadded;
+use elfo_utils::{CachePadded, RateLimiter};
 
 #[allow(unreachable_pub)] // Actually, it's reachable via `elfo::_priv`.
 pub use self::{dump_item::*, sequence_no::SequenceNo};
 
+pub use hider::hide;
+pub(crate) use hider::set_in_dumping;
+
 use self::sequence_no::SequenceNoGenerator;
 use crate::{
     envelope,
@@ -28,6 +31,7 @@ use crate::{
 };
 
 mod dump_item;
+mod hider;
 mod sequence_no;
 
 const SHARD_COUNT: usize = 16;
@@ -40,15 +44,26 @@ thread_local! {
 
 // Reexported in `elfo::_priv`.
 #[derive(Clone, Default)]
+#[doc(hidden)]
 pub struct Dumper {
     per_system: Arc<PerSystem>,
     per_group: Arc<PerGroup>,
 }
 
-#[derive(Default, Deserialize)]
+#[derive(Deserialize)]
 #[serde(default)]
 pub(crate) struct DumpingConfig {
     disabled: bool,
+    max_rate: u64,
+}
+
+impl Default for DumpingConfig {
+    fn default() -> Self {
+        Self {
+            disabled: false,
+            max_rate: 100_000,
+        }
+    }
 }
 
 #[derive(Default)]
@@ -58,8 +73,9 @@ struct PerSystem {
 
 #[derive(Default)]
 struct PerGroup {
-    // TODO: add a rate limiter.
-    sequence_no_gen: CachePadded<SequenceNoGenerator>,
+    // TODO: add `CachePadded`.
+    rate_limiter: RateLimiter,
+    sequence_no_gen: SequenceNoGenerator,
     is_possible: bool,
     is_disabled: AtomicBool,
 }
@@ -68,7 +84,8 @@ impl Dumper {
     pub(crate) fn for_group(&self, is_possible: bool) -> Self {
         let mut specific = self.clone();
         specific.per_group = Arc::new(PerGroup {
-            sequence_no_gen: CachePadded(SequenceNoGenerator::default()),
+            rate_limiter: RateLimiter::default(),
+            sequence_no_gen: SequenceNoGenerator::default(),
             is_possible,
             is_disabled: AtomicBool::new(false),
         });
@@ -79,11 +96,22 @@ impl Dumper {
         self.per_group
             .is_disabled
             .store(config.disabled, Ordering::Relaxed);
+        self.per_group.rate_limiter.configure(config.max_rate);
     }
 
     #[inline]
     pub fn is_enabled(&self) -> bool {
-        self.per_group.is_possible && !self.per_group.is_disabled.load(Ordering::Relaxed)
+        if !self.per_group.is_possible || self.per_group.is_disabled.load(Ordering::Relaxed) {
+            return false;
+        }
+
+        if self.per_group.rate_limiter.acquire() {
+            // TODO: increase `sequence_no` anyway.
+            true
+        } else {
+            increment_counter!("elfo_limited_dumps_total");
+            false
+        }
     }
 
     #[inline(always)]
@@ -155,6 +183,8 @@ impl Dumper {
             return;
         }
 
+        increment_counter!("elfo_emitted_dumps_total");
+
         queue.push_back(item);
     }
 
@@ -169,6 +199,7 @@ impl Dumper {
 }
 
 // Reexported in `elfo::_priv`.
+#[doc(hidden)]
 pub struct Drain<'a> {
     dumper: &'a Dumper,
     shard_no: usize,
@@ -280,7 +311,13 @@ mod tests {
             assert!(drain.next().is_none());
         };
 
-        let scope = Scope::new(Addr::NULL, Addr::NULL, meta.clone(), Default::default());
+        let scope = Scope::new(
+            Addr::NULL,
+            Addr::NULL,
+            meta.clone(),
+            Default::default(),
+            Default::default(),
+        );
         scope.set_trace_id(trace_id);
         scope.within(f).await;
     }
